"""The generic configuration file in a .fmu directory."""

from __future__ import annotations

import copy
import getpass
from datetime import UTC, datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Final, Self

from fmu.settings._logging import null_logger
from fmu.settings.models.project_config import ProjectConfig
from fmu.settings.models.user_config import UserConfig

from .pydantic_resource_manager import (
    MutablePydanticResourceManager,
)

if TYPE_CHECKING:
    from collections.abc import Mapping

    # Avoid circular dependency for type hint in __init__ only
    from fmu.settings._fmu_dir import (
        ProjectFMUDirectory,
        UserFMUDirectory,
    )

logger: Final = null_logger(__name__)


class ProjectConfigManager(MutablePydanticResourceManager[ProjectConfig]):
    """Manages the .fmu configuration file in a project."""

    fmu_dir: ProjectFMUDirectory

    def __init__(self: Self, fmu_dir: ProjectFMUDirectory) -> None:
        """Initializes the ProjectConfig resource manager."""
        super().__init__(fmu_dir, ProjectConfig)

    @property
    def relative_path(self: Self) -> Path:
        """Returns the relative path to the config file."""
        return Path("config.json")

    @property
    def diff_ignore_fields(self: Self) -> list[str]:
        """Returns a list of all config fields that should be ignored in a diff."""
        return ["created_at", "created_by", "last_modified_at", "last_modified_by"]

    @property
    def diff_list_keys(self: Self) -> Mapping[str, str]:
        """List field identity keys used for per-item diffing."""
        return {
            "rms.zones": "name",
            "rms.horizons": "name",
            "rms.wells": "name",
            "masterdata.smda.country": "uuid",
            "masterdata.smda.discovery": "uuid",
            "masterdata.smda.field": "uuid",
        }

    def save(self: Self, model: ProjectConfig) -> None:
        """Save the ProjectConfig to disk, updating last_modified fields."""
        model_dict = model.model_dump()
        model_dict["last_modified_at"] = datetime.now(UTC)
        model_dict["last_modified_by"] = getpass.getuser()
        updated_model = ProjectConfig.model_validate(model_dict)
        super().save(updated_model)

    def set(self: Self, key: str, value: Any) -> None:
        """Sets a project config value by key and writes to changelog."""
        old_resource_dict = copy.deepcopy(self.load().model_dump())
        super().set(key, value)
        self.fmu_dir.changelog.log_update_to_changelog(
            updates={key: value},
            old_resource_dict=old_resource_dict,
            relative_path=self.relative_path,
        )

    def update(self: Self, updates: dict[str, Any]) -> ProjectConfig:
        """Updates project config values and writes to changelog."""
        old_resource_dict = copy.deepcopy(self.load().model_dump())
        updated_resource = super().update(updates)
        self.fmu_dir.changelog.log_update_to_changelog(
            updates=updates,
            old_resource_dict=old_resource_dict,
            relative_path=self.relative_path,
        )
        return updated_resource


class UserConfigManager(MutablePydanticResourceManager[UserConfig]):
    """Manages the .fmu configuration file in a user's home directory."""

    def __init__(self: Self, fmu_dir: UserFMUDirectory) -> None:
        """Initializes the UserConfig resource manager."""
        super().__init__(fmu_dir, UserConfig)

    @property
    def relative_path(self: Self) -> Path:
        """Returns the relative path to the config file."""
        return Path("config.json")

    def save(self: Self, model: UserConfig) -> None:
        """Save the UserConfig to disk, updating last_modified_at."""
        model_dict = model.model_dump()
        model_dict["last_modified_at"] = datetime.now(UTC)
        updated_model = UserConfig.model_validate(model_dict)
        super().save(updated_model)



"""Tests for mapping models."""

from uuid import UUID, uuid4

import pytest
from fmu.datamodels.context.mappings import (
    DataSystem,
    MappingType,
    RelationType,
    StratigraphyIdentifierMapping,
)
from pydantic import ValidationError

from fmu.settings.models.mappings import MappingGroup


def _make_mapping(  # noqa: PLR0913
    source_id: str,
    target_id: str,
    relation_type: RelationType,
    source_system: DataSystem = DataSystem.rms,
    target_system: DataSystem = DataSystem.smda,
    target_uuid: UUID | None = None,
) -> StratigraphyIdentifierMapping:
    return StratigraphyIdentifierMapping(
        source_system=source_system,
        target_system=target_system,
        relation_type=relation_type,
        source_id=source_id,
        target_id=target_id,
        target_uuid=target_uuid,
    )


def test_mapping_group_count_mappings_by_relation_type() -> None:
    """Tests that the count of mappings with the given relation_type is returned."""
    target_id = "Viking GP."
    primary = _make_mapping("Viking Gp", target_id, RelationType.primary)
    alias = _make_mapping("Viking gp", target_id, RelationType.alias)
    alias_two = _make_mapping("VIKING GP", target_id, RelationType.alias)
    alias_three = _make_mapping("viking.gp", target_id, RelationType.alias)
    equivalent = _make_mapping("Viking GP.", target_id, RelationType.equivalent)

    mapping_group = MappingGroup(
        target_id=target_id,
        mapping_type=MappingType.stratigraphy,
        target_system=DataSystem.smda,
        source_system=DataSystem.rms,
        mappings=[primary, alias, alias_two, alias_three, equivalent],
    )

    assert mapping_group._count_mappings_by_relation_type(RelationType.primary) == 1
    assert mapping_group._count_mappings_by_relation_type(RelationType.equivalent) == 1
    expected_alias_mappings = 3
    assert (
        mapping_group._count_mappings_by_relation_type(RelationType.alias)
        == expected_alias_mappings
    )


def test_mapping_group_serializes_without_system_and_target_fields() -> None:
    """MappingGroup serializes mappings without redundant fields."""
    target_id = "Viking GP."
    target_uuid = uuid4()
    primary = _make_mapping(
        "Viking Gp", target_id, RelationType.primary, target_uuid=target_uuid
    )
    alias = _make_mapping(
        "Viking Group", target_id, RelationType.alias, target_uuid=target_uuid
    )
    equivalent = _make_mapping(
        target_id, target_id, RelationType.equivalent, target_uuid=target_uuid
    )

    group = MappingGroup(
        target_id=target_id,
        target_uuid=target_uuid,
        mapping_type=MappingType.stratigraphy,
        target_system=DataSystem.smda,
        source_system=DataSystem.rms,
        mappings=[primary, alias, equivalent],
    )

    display_dict = group.model_dump()
    assert display_dict["official_name"] == target_id
    assert display_dict["target_uuid"] == target_uuid
    assert display_dict["mapping_type"] == "stratigraphy"
    assert display_dict["target_system"] == "smda"
    assert display_dict["source_system"] == "rms"
    for mapping in display_dict["mappings"]:
        assert "source_system" not in mapping
        assert "target_system" not in mapping
        assert "mapping_type" not in mapping
        assert "target_id" not in mapping
        assert "target_uuid" not in mapping
        assert "relation_type" in mapping
        assert "source_id" in mapping


def test_mapping_group_rejects_multiple_primary_mappings() -> None:
    """MappingGroup with more than one primary mapping raises ValidationError."""
    target_id = "Viking GP."
    primary = _make_mapping("Viking Gp", target_id, RelationType.primary)
    primary_two = _make_mapping("Viking Gp 2", target_id, RelationType.primary)
    with pytest.raises(ValidationError, match="at most one primary mapping."):
        MappingGroup(
            target_id=target_id,
            mapping_type=MappingType.stratigraphy,
            target_system=DataSystem.smda,
            source_system=DataSystem.rms,
            mappings=[primary, primary_two],
        )


def test_mapping_group_rejects_multiple_equivalent_mappings() -> None:
    """MappingGroup with more than one equivalent mapping raises ValidationError."""
    target_id = "Viking GP."
    equivalent = _make_mapping("Viking GP.", target_id, RelationType.equivalent)
    equivalent_two = _make_mapping("Viking GP.", target_id, RelationType.equivalent)
    with pytest.raises(ValidationError, match="at most one equivalent mapping."):
        MappingGroup(
            target_id=target_id,
            mapping_type=MappingType.stratigraphy,
            target_system=DataSystem.smda,
            source_system=DataSystem.rms,
            mappings=[equivalent, equivalent_two],
        )


def test_mapping_group_with_alias_requires_primary_relation() -> None:
    """MappingsGroup with alias mappings and no primary mapping raises ValidationError.

    A MappingGroup containing alias mappings will raise a ValidationError
    Error if no primary mapping exists.
    """
    target_id = "Viking GP."
    alias = _make_mapping("Viking gp", target_id, RelationType.alias)
    alias_two = _make_mapping("VIKING GP", target_id, RelationType.alias)
    equivalent = _make_mapping("Viking GP.", target_id, RelationType.equivalent)
    with pytest.raises(
        ValidationError, match="contains alias relations but no primary relation."
    ):
        MappingGroup(
            target_id=target_id,
            mapping_type=MappingType.stratigraphy,
            target_system=DataSystem.smda,
            source_system=DataSystem.rms,
            mappings=[alias, alias_two, equivalent],
        )


def test_mapping_group_requires_no_duplicate_mappings() -> None:
    """MappingGroup with duplicate mappings raises ValidationError."""
    target_id = "Viking GP."
    primary = _make_mapping("VIKING GP", target_id, RelationType.primary)
    alias = _make_mapping("Viking gp", target_id, RelationType.alias)
    alias_two = _make_mapping("Viking gp", target_id, RelationType.alias)
    with pytest.raises(ValidationError, match="Duplicate mapping in group:"):
        MappingGroup(
            target_id=target_id,
            mapping_type=MappingType.stratigraphy,
            target_system=DataSystem.smda,
            source_system=DataSystem.rms,
            mappings=[primary, alias, alias_two],
        )


def test_mapping_group_requires_shared_target_id() -> None:
    """All mappings must share the group target_id."""
    primary = _make_mapping("Viking Gp", "Viking GP.", RelationType.primary)
    different_target = _make_mapping("westeros", "Westeros", RelationType.alias)

    with pytest.raises(ValidationError, match="target_id"):
        MappingGroup(
            target_id="Viking GP.",
            mapping_type=MappingType.stratigraphy,
            target_system=DataSystem.smda,
            source_system=DataSystem.rms,
            mappings=[primary, different_target],
        )


def test_mapping_group_requires_shared_target_system() -> None:
    """All mappings must share the group target_system."""
    primary = _make_mapping("Viking Gp", "Viking GP.", RelationType.primary)
    different_target_system = _make_mapping(
        "Viking Gp 2",
        "Viking GP.",
        RelationType.alias,
        target_system=DataSystem.fmu,
    )

    with pytest.raises(ValidationError, match="target_system"):
        MappingGroup(
            target_id="Viking GP.",
            mapping_type=MappingType.stratigraphy,
            target_system=DataSystem.smda,
            source_system=DataSystem.rms,
            mappings=[primary, different_target_system],
        )


def test_mapping_group_requires_shared_source_system() -> None:
    """All mappings must share the group source_system."""
    primary = _make_mapping("Viking Gp", "Viking GP.", RelationType.primary)
    different_source_system = _make_mapping(
        "Viking Gp 2",
        "Viking GP.",
        RelationType.alias,
        source_system=DataSystem.fmu,
    )

    with pytest.raises(ValidationError, match="source_system"):
        MappingGroup(
            target_id="Viking GP.",
            mapping_type=MappingType.stratigraphy,
            target_system=DataSystem.smda,
            source_system=DataSystem.rms,
            mappings=[primary, different_source_system],
        )



"""Diff models shared by resource managers and API consumers."""

from __future__ import annotations

from typing import Any

from pydantic import BaseModel


class ScalarFieldDiff(BaseModel):
    """Diff entry for non-list fields."""

    field_path: str
    before: Any
    after: Any


class ListUpdatedEntry(BaseModel):
    """Before and after values for an updated list item."""

    key: Any
    before: dict[str, Any]
    after: dict[str, Any]


class ListFieldDiff(BaseModel):
    """Diff entry for list fields with per-item changes."""

    field_path: str
    added: list[dict[str, Any]]
    removed: list[dict[str, Any]]
    updated: list[ListUpdatedEntry]


ResourceDiff = ScalarFieldDiff | ListFieldDiff



"""Model for the mappings.json file."""

from typing import Any, Self
from uuid import UUID

from pydantic import BaseModel, Field, model_serializer, model_validator

from fmu.datamodels.context.mappings import (
    AnyIdentifierMapping,
    DataSystem,
    MappingType,
    RelationType,
    StratigraphyMappings,
)


class Mappings(BaseModel):
    """Represents the mappings file in a .fmu directory."""

    stratigraphy: StratigraphyMappings = Field(
        default_factory=lambda: StratigraphyMappings(root=[])
    )
    """Stratigraphy mappings in the mappings file."""

    # Todo: Add wells model
    wells: list[Any] = Field(default_factory=list)
    """Well mappings in the mappings file."""


class MappingGroup(BaseModel):
    """A mapping group containing related mappings with a shared target context.

    This is a view of the data for GUI display purposes, not how it's stored.
    Groups mappings that share target_id, target_uuid (if set), mapping_type,
    target_system, and source_system.
    """

    target_id: str
    target_uuid: UUID | None = None
    mapping_type: MappingType
    target_system: DataSystem
    source_system: DataSystem
    mappings: list[AnyIdentifierMapping]

    def _count_mappings_by_relation_type(
        self: Self, relation_type: RelationType
    ) -> int:
        return sum(
            1 for mapping in self.mappings if mapping.relation_type == relation_type
        )

    @model_validator(mode="after")
    def validate_group(self) -> "MappingGroup":
        """Ensure MappingGroup contains a valid combination of mappings.

        Validates that:
        - At most one primary mapping per mapping group
        - At most one equivalent mapping per mapping group
        - Mapping group with alias mapping(s) requires a primary mapping
        - No duplicate mappings in a mappping group
        - All mappings in a mapping group share the same target_id, mapping_type,
          target_system and source_system

        Raises:
            ValueError: If validation fails or mapping type is unsupported
        """
        if not self.mappings:
            return self

        primary_mappings_count = self._count_mappings_by_relation_type(
            RelationType.primary
        )
        has_alias_mapping = (
            self._count_mappings_by_relation_type(RelationType.alias) > 0
        )

        if primary_mappings_count > 1:
            raise ValueError(
                f"MappingGroup for target '{self.target_id}' must contain at most one "
                "primary mapping."
            )

        if self._count_mappings_by_relation_type(RelationType.equivalent) > 1:
            raise ValueError(
                f"MappingGroup for target '{self.target_id}' must contain at most one "
                "equivalent mapping."
            )

        if has_alias_mapping and primary_mappings_count == 0:
            raise ValueError(
                f"MappingGroup for target '{self.target_id}' contains "
                "alias relations but no primary relation. Aliases require "
                "a primary unofficial identifier."
            )

        seen: set[tuple[str, UUID | None, RelationType]] = set()
        for mapping in self.mappings:
            mapping_key = (
                mapping.source_id,
                mapping.source_uuid,
                mapping.relation_type,
            )
            if mapping_key in seen:
                raise ValueError(
                    f"Duplicate mapping in group: source_id='{mapping.source_id}', "
                    f"source_uuid='{mapping.source_uuid}', "
                    f"relation_type='{mapping.relation_type}'."
                )

            if mapping.target_id != self.target_id:
                raise ValueError(
                    "All mappings in MappingGroup must share target_id "
                    f"'{self.target_id}'."
                )
            if mapping.mapping_type != self.mapping_type:
                raise ValueError(
                    "All mappings in MappingGroup must share mapping_type "
                    f"'{self.mapping_type}'."
                )
            if mapping.target_system != self.target_system:
                raise ValueError(
                    "All mappings in MappingGroup must share target_system "
                    f"'{self.target_system}'."
                )
            if mapping.source_system != self.source_system:
                raise ValueError(
                    "All mappings in MappingGroup must share source_system "
                    f"'{self.source_system}'."
                )
            if (
                mapping.target_uuid is not None
                and self.target_uuid is not None
                and mapping.target_uuid != self.target_uuid
            ):
                raise ValueError(
                    "All mappings in MappingGroup must share target_uuid "
                    f"'{self.target_uuid}'."
                )

            seen.add(mapping_key)

        return self

    @model_serializer(mode="plain")
    def serialize_for_display(self) -> dict[str, Any]:
        """Serialize to the display schema for API responses."""
        excluded_fields = {
            "source_system",
            "target_system",
            "mapping_type",
            "target_id",
            "target_uuid",
        }
        return {
            "official_name": self.target_id,
            "target_uuid": self.target_uuid,
            "mapping_type": self.mapping_type.value,
            "target_system": self.target_system.value,
            "source_system": self.source_system.value,
            "mappings": [
                {
                    key: value
                    for key, value in mapping.model_dump().items()
                    if key not in excluded_fields
                }
                for mapping in self.mappings
            ],
        }



"""Contains models used in this package.

Some models contained here may also be used outside this
package.
"""

from .diff import (
    ListFieldDiff,
    ListUpdatedEntry,
    ResourceDiff,
    ScalarFieldDiff,
)

__all__ = [
    "ListFieldDiff",
    "ListUpdatedEntry",
    "ResourceDiff",
    "ScalarFieldDiff",
]
